/*

    This program is free software; you can redistribute it and/or
    modify it under the terms of the GNU General Public License as
    published by the Free Software Foundation; either version 3 of the
    License, or (at your option) any later version.

    This program is distributed in the hope that it will be useful,
    but WITHOUT ANY WARRANTY; without even the implied warranty of
    MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the GNU
    General Public License for more details.

    You should have received a copy of the GNU General Public License
    along with this program; if not, write to the Free Software
    Foundation, Inc., 51 Franklin St, Fifth Floor, Boston, MA
    02110-1301 USA

    Copyright Alexandre Becoulet <alexandre.becoulet@free.fr> (C) 2009

*/

#ifndef _GCT_ATOMIC_PTHREAD_H_
#define _GCT_ATOMIC_PTHREAD_H_

#if !(defined(_REENTRANT) || defined(_THREAD_SAFE)) && !defined(GCT_CONFIG_NOWARNING)
# warning pthread requires _REENTRANT or _THREAD_SAFE to be defined
#endif

/* Use pthread spinlock */

#if defined (GCT_CONFIG_ATOMIC_PTHREAD_SPIN)

# if (_XOPEN_SOURCE < 600) && !defined(GCT_CONFIG_NOWARNING)
#  warning pthread spinlock requires _XOPEN_SOURCE >= 600
# endif

# define _GCT_ATOMIC_LOCKTYPE pthread_spinlock_t
# define _GCT_ATOMIC_LOCK(lock) \
  { gct_err_t err = pthread_spin_lock(lock); _GCT_ASSERT(!err); }
# define _GCT_ATOMIC_UNLOCK(lock) \
  { gct_err_t err = pthread_spin_unlock(lock); _GCT_ASSERT(!err); }

/* Use pthread mutex */

#elif defined (GCT_CONFIG_ATOMIC_PTHREAD_MUTEX)

# define _GCT_ATOMIC_LOCKTYPE pthread_mutex_t
# define _GCT_ATOMIC_LOCK(lock) \
  { gct_err_t err = pthread_mutex_lock(lock); _GCT_ASSERT(!err); }
# define _GCT_ATOMIC_UNLOCK(lock) \
  { gct_err_t err = pthread_mutex_unlock(lock); _GCT_ASSERT(!err); }

#else
# error No atomic pthread locking mechanism selected
#endif

#include <pthread.h>

typedef int_fast8_t     _gct_atomic_int_t;

typedef struct  _gct_atomic_s
{
  volatile _gct_atomic_int_t val;
  _GCT_ATOMIC_LOCKTYPE lock;
}               _gct_atomic_t;

/* static value init */
#ifdef GCT_CONFIG_ATOMIC_PTHREAD_SPIN
# define _GCT_ATOMIC_INITIALIZER(v)     { .val = (v), .lock = PTHREAD_SPINLOCK_INITIALIZER }
#else
# define _GCT_ATOMIC_INITIALIZER(v)     { .val = (v), .lock = PTHREAD_MUTEX_INITIALIZER }
#endif

/* init atomic value lock if needed */
_GCT_INTERNAL void
_gct_atomic_init(_gct_atomic_t *a)
{
  gct_err_t err;
#ifdef GCT_CONFIG_ATOMIC_PTHREAD_SPIN
  err = pthread_spin_init(&a->lock, PTHREAD_PROCESS_PRIVATE);
#else
  err = pthread_mutex_init(&a->lock, NULL);
#endif
  _GCT_ASSERT(!err);
}

/* destroy atomic value lock if needed */
_GCT_INTERNAL void
_gct_atomic_destroy(_gct_atomic_t *a)
{
  gct_err_t err;
#ifdef GCT_CONFIG_ATOMIC_PTHREAD_SPIN
  err = pthread_spin_destroy(&a->lock);
#else
  err = pthread_mutex_destroy(&a->lock);
#endif
  _GCT_ASSERT(!err);
}

/* set atomic value */
_GCT_INTERNAL void
_gct_atomic_set(_gct_atomic_t *a, _gct_atomic_int_t v)
{
  a->val = v;
}

/* get atomic value */
_GCT_INTERNAL _gct_atomic_int_t
_gct_atomic_get(_gct_atomic_t *a)
{
  return a->val;
}

/* increment atomic value return true if not 0 */
_GCT_INTERNAL gct_bool_t
_gct_atomic_inc(_gct_atomic_t *a)
{
  _gct_atomic_int_t     res;

  _GCT_ATOMIC_LOCK(&a->lock);
  res = ++a->val;
  _GCT_ATOMIC_UNLOCK(&a->lock);

  return res != 0;
}

/* decrement atomic value return true if not 0 */
_GCT_INTERNAL gct_bool_t
_gct_atomic_dec(_gct_atomic_t *a)
{
  _gct_atomic_int_t     res;

  _GCT_ATOMIC_LOCK(&a->lock);
  res = --a->val;
  _GCT_ATOMIC_UNLOCK(&a->lock);

  return res != 0;
}

/* bit test */
_GCT_INTERNAL gct_bool_t
_gct_atomic_bit_test(_gct_atomic_t *a, _gct_atomic_int_t n)
{
  return (a->val & (1 << n)) != 0;
}

/* bit test and set */
_GCT_INTERNAL gct_bool_t
_gct_atomic_bit_test_set(_gct_atomic_t *a, _gct_atomic_int_t n)
{
  _gct_atomic_int_t     res;
  _gct_atomic_int_t     mask = (1 << n);

  _GCT_ATOMIC_LOCK(&a->lock);
  res = a->val & mask;
  a->val |= mask;
  _GCT_ATOMIC_UNLOCK(&a->lock);

  return res != 0;
}

/* bit test and clear */
_GCT_INTERNAL gct_bool_t
_gct_atomic_bit_test_clr(_gct_atomic_t *a, _gct_atomic_int_t n)
{
  _gct_atomic_int_t     res;
  _gct_atomic_int_t     mask = (1 << n);

  _GCT_ATOMIC_LOCK(&a->lock);
  res = a->val & mask;
  a->val &= ~mask;
  _GCT_ATOMIC_UNLOCK(&a->lock);

  return res != 0;
}

/* bit clear */
_GCT_INTERNAL void
_gct_atomic_bit_clr(_gct_atomic_t *a, _gct_atomic_int_t n)
{
  _GCT_ATOMIC_LOCK(&a->lock);
  a->val &= ~(1 << n);
  _GCT_ATOMIC_UNLOCK(&a->lock);
}

/* bit set */
_GCT_INTERNAL void
_gct_atomic_bit_set(_gct_atomic_t *a, _gct_atomic_int_t n)
{
  _GCT_ATOMIC_LOCK(&a->lock);
  a->val |= (1 << n);
  _GCT_ATOMIC_UNLOCK(&a->lock);
}

#endif

