/* -*- c -*-

    This program is free software; you can redistribute it and/or
    modify it under the terms of the GNU General Public License as
    published by the Free Software Foundation; either version 3 of the
    License, or (at your option) any later version.

    This program is distributed in the hope that it will be useful,
    but WITHOUT ANY WARRANTY; without even the implied warranty of
    MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the GNU
    General Public License for more details.

    You should have received a copy of the GNU General Public License
    along with this program; if not, write to the Free Software
    Foundation, Inc., 51 Franklin St, Fifth Floor, Boston, MA
    02110-1301 USA

    Copyright Alexandre Becoulet <alexandre.becoulet@free.fr> (C) 2009

*/

#ifndef _GCT_OBJECT_H_
#define _GCT_OBJECT_H_

#include <stdarg.h>

#include <gct/_platform.h>

#define GCT_OBJECT_STORAGE_FREE(fcn) void (fcn)(void * storage)
typedef GCT_OBJECT_STORAGE_FREE(object_storage_free_t);

/* backslash-region-begin */
#define GCT_OBJECT_TYPES(name, algo, type)

typedef type * name##_object_t;
typedef type name##_object_base_t;

typedef struct name##_entry_s
{
  object_storage_free_t *storage_free;
  _GCT_OBJ_##algo##_ENTRY_FIELDS(name);
} name##_entry_t;
/* backslash-region-end */


/***********************************************************************
 *      user defined object related functions prototypes
 */

/**
   Object contructor function provided by user.

   @param obj allocated object ready for initialization.
   @param ap stdarg object used to access extra parameters passed to new()

   Some gct related fields may already be initialized and should not
   be overwritten with a global memset operation.
*/
#define GCT_OBJECT_CONSTRUCTOR(name)                                \
gct_err_t                                                    \
(name##_constructor)(name##_object_t obj,                       \
                     va_list ap)

/**
   Object destructor called before freeing object storage area
 */
#define GCT_OBJECT_DESTRUCTOR(name)                                 \
void                                                            \
(name##_destructor)(name##_object_t obj)


/***********************************************************************
 *      object access functions prototypes
 */


/**
   Create a new object by allocating storage area and calling its
   constrctor. Object ref count is set to 1 if any.

   @param storage pointer to user provided storage area for
   object. Use NULL pointer to let gct allocate storage for the
   object. if a storage area pointer is specified, no automatic
   storage freeing will occure after calling object destructor.

   @return constructor result, NULL if failed
 */
#define _GCT_OBJ_NEW_PROTOTYPE(name, prefix)                        \
name##_object_t                                                 \
prefix##_new    (void *storage, ...)

/**
   Create a new object by calling its constrctor. Object ref count is
   set to 1 if any.

   @param storage pointer to storage area for object
   @param free pointer to storage freeing function.
   @return constructor result, NULL if failed
*/
#define _GCT_OBJ_CONSTRUCT_PROTOTYPE(name, prefix)                  \
gct_err_t                                                    \
prefix##_construct(void *storage,                               \
                   object_storage_free_t *sfree,                \
                   va_list ap)

/**
   Mark an object for deletion if available depending on deletion
   policy.
*/
#define _GCT_OBJ_DELETE_PROTOTYPE(name, prefix)                     \
gct_err_t                                                    \
prefix##_delete(name##_object_t object)

/**
   Get a new reference to an object. May fail if _object.has been
   marked for deletion.

   @return pointer to object, NULL if failed
*/
#define _GCT_OBJ_REFNEW_PROTOTYPE(name, prefix)                     \
name##_object_t                                                 \
prefix##_refnew (name##_object_t object)

/**
   Drop a reference to an object. Destrctor will be called if
   reference count if set to one.
*/
#define _GCT_OBJ_REFDROP_PROTOTYPE(name, prefix)                    \
void                                                            \
prefix##_refdrop(name##_object_t object)

/**
   Get reference conter value
*/
#define _GCT_OBJ_REFCOUNT_PROTOTYPE(name, prefix)                   \
size_t                                                          \
prefix##_refcount(name##_object_t object)

/* backslash-region-begin */
#define GCT_OBJECT_PROTOTYPES(name, attr, prefix)

_GCT_UNUSED attr _GCT_OBJ_NEW_PROTOTYPE(name, prefix);
_GCT_UNUSED attr _GCT_OBJ_DELETE_PROTOTYPE(name, prefix);
_GCT_UNUSED attr _GCT_OBJ_REFNEW_PROTOTYPE(name, prefix);
_GCT_UNUSED attr _GCT_OBJ_REFDROP_PROTOTYPE(name, prefix);
_GCT_UNUSED attr _GCT_OBJ_REFCOUNT_PROTOTYPE(name, prefix);
/* backslash-region-end */

#define GCT_OBJECT_FCNS(name, algo, attr, prefix, ...)              \
 _GCT_OBJ_##algo##_FCNS(attr, name, prefix, __VA_ARGS__)

#define GCT_OBJECT_INITIALIZER(name, algo)          _GCT_OBJ_##algo##_INITIALIZER

#endif

